import os
from therix.services.therix_trace_handler import TherixTraceHandler



therix_trace_handler = TherixTraceHandler(
    therix_api_key = os.getenv('THERIX_API_KEY'),
    agent_id = "some-agent-id",
    session_id= "some-session-id",
    metadata = {
        "metadata": "some-metadata"
    } 
)


handler = therix_trace_handler.get_handler()







import json
import os
from langfuse.callback import CallbackHandler
from pydantic import Json
from therix.core.version import get_current_version
from therix.core.constants import API_Endpoint


class TherixTraceHandler:
    def __init__(self, therix_api_key ,agent_id=None, session_id=None, metadata=None):
        self.agent_id = agent_id
        self.session_id = session_id
        self.metadata = metadata
        self.therix_api_key = therix_api_key

        # Initialize the internal handler
        self.handler = self._initialize_handler()

    def _initialize_handler(self):

            handler_args = {
                "secret_key": 'some-secret-key',
                "public_key":self.therix_api_key ,
                "host": API_Endpoint.TRACE_INGESTION,
                "trace_name": self.agent_id,
                "session_id": self.session_id,
                "version": get_current_version(),
                "metadata": self.metadata
            }
            

            return CallbackHandler(**handler_args)

      
    def get_handler(self):
        return self.handler





    


from typing import List, Text
from pydantic import BaseModel, Field
from therix.core.data_sources import PDFDataSource
from therix.core.embedding_models import BedrockTitanEmbedding

from therix.core.output_parser import OutputParserWrapper
from therix.core.agent import Agent

from therix.core.system_prompt_config import SystemPromptConfig
from therix.core.agent import Agent
from therix.core.data_sources import TextDataSource
from therix.core.embedding_models import BedrockTitanEmbedding
from therix.core.inference_models import GroqMixtral87bInferenceModel,GroqLlama370b
import asyncio

GROQ_API_KEY=''

sys_prompt = """Answer the question like a pirate based only on the following context and reply with your capabilities if something is out of context.
        Context: 
        {{context}}

        Question: {{question}}
        
        Give the output as follows : {{format_instructions}}

        """
        # Always adress me with my name {name}


variables = {
        "name": "Abhishek Dubey",
}

agent = Agent(name="new smarttttttt agent")
(
        agent.add(TextDataSource(config={"files": ["../../test-data/rat.pdf"]}))   
        .add(BedrockTitanEmbedding(config={"bedrock_aws_access_key_id" : "",
                                "bedrock_aws_secret_access_key" : "",
                                "bedrock_aws_session_token" : "",
                                "bedrock_region_name" : "us-east-1"}))
        .add(GroqLlama370b(config={"groq_api_key": GROQ_API_KEY}))
        # .add(SystemPromptConfig(config={"system_prompt" : "new-prompt"}))
        .save()
    )

agent.preprocess_data()

print(agent.id)

class TestDetails(BaseModel):
        name: str = Field(description="Name of the Topic")
        description: str = Field(description="Short description of the Topic")
        citations: str = Field(description="add source of every topic, from where it is generated")
        page: str = Field(description="page number of the topic")


class OutputParserJSON(BaseModel):
    tests: List[TestDetails] = Field(description="Topic")
    
answer = agent.invoke("What kind of experiments are performed in the study?" , output_parser=OutputParserWrapper.parse_output(pydantic_object=OutputParserJSON), prompt_variables=variables)

print(answer)


import asyncio
from enum import Enum
import json
from therix.utils.rag import get_inference_model
from langchain.chains.llm import LLMChain
from langchain_core.prompts import PromptTemplate
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.text_splitter import CharacterTextSplitter
from langchain.docstore.document import Document
from langchain.chains import MapReduceDocumentsChain, ReduceDocumentsChain
from langchain_core.output_parsers import JsonOutputParser
from therix.services.trace_handler import get_trace_handler


class SummarizerTypeMaster(Enum):
    EXTRACTIVE = 'EXTRACTIVE'
    ABSTRACTIVE = 'ABSTRACTIVE'

PROMPT_TEMPLATE = {
    "EXTRACTIVE": """
        Summarize the provided context below:

        {context}

        ---

        Craft your response with conciseness and accuracy, including only the information provided in the context. 
        Use null values for any missing information.

        Please structure your response in the following JSON format:
        {response_schema_json}
    """,
    "ABSTRACTIVE": "Provide a concise summary for the following text using abstractive summarization:\n\n{context}."
}




async def async_summarizer(summarizer_config, inference_model_details, text, trace_details, pipeline_id, system_prompt=None, session_id=None, trace_api=None,metadata=None, config = None):

    if not config:
        therix_trace_handler = await get_trace_handler(trace_details, trace_api, system_prompt, pipeline_id, session_id,metadata=metadata)
        chain_callbacks = [therix_trace_handler] if therix_trace_handler else []
    else:
        therix_trace_handler = config    

    pydantic_prompt = summarizer_config['pydantic_model']
    summarization_type = summarizer_config['summarization_type']
    llm = get_inference_model(inference_model_details.name, inference_model_details.config)

    if summarization_type == 'ABSTRACTIVE':
        map_template = PromptTemplate.from_template(
            """Read the following document and provide a detailed summary of the document.\n{pages}"""
        )
        reduce_template = """The following is a set of summaries:\n{pages}\nTake these and distill them into a final, consolidated summary. Verify each and every detail and make sure that the final summary is accurate and complete."""
        
        map_chain = LLMChain(llm=llm, prompt=map_template)
        reduce_chain = LLMChain(llm=llm, prompt=reduce_template)
        
        combine_documents_chain = StuffDocumentsChain(
            llm_chain=reduce_chain, document_variable_name="pages"
        )

        reduce_documents_chain = ReduceDocumentsChain(
            combine_documents_chain=combine_documents_chain,
            collapse_documents_chain=combine_documents_chain,
            token_max=4000,
        )

        map_reduce_chain = MapReduceDocumentsChain(
            llm_chain=map_chain,
            reduce_documents_chain=reduce_documents_chain,
            document_variable_name="pages",
            return_intermediate_steps=False,
        )

        docs = Document(page_content=text, metadata={"source": "local"})
        text_splitter = CharacterTextSplitter.from_tiktoken_encoder(chunk_size=1000, chunk_overlap=0)
        split_docs = text_splitter.split_documents([docs])

        summary = await map_reduce_chain.arun(split_docs)
    else:
        summary = text

    parser = JsonOutputParser()
    
    prompt_template = system_prompt.get("system_prompt") if system_prompt and "system_prompt" in system_prompt else system_prompt
    prompt = PromptTemplate(
        template=prompt_template or PROMPT_TEMPLATE[summarization_type],
        input_variables=["context"],
        partial_variables={"response_schema_json": pydantic_prompt},
    )

    if summarization_type == SummarizerTypeMaster.EXTRACTIVE.value:
        chain = prompt | llm | parser
        result = await chain.ainvoke({"context": summary}, config={"callbacks": [therix_trace_handler]})
    else:
        chain = prompt | llm
        result = await chain.ainvoke({"context": summary}, config={"callbacks": [therix_trace_handler]})

    return json.dumps(result) if summarization_type == SummarizerTypeMaster.EXTRACTIVE.value else result.content

def summarizer(summarizer_config, inference_model_details, text, trace_details, pipeline_id, system_prompt=None, session_id=None, trace_api=None,metadata=None, config = None):
    if not config:
        therix_trace_handler = get_trace_handler(trace_details, trace_api, system_prompt, pipeline_id, session_id, metadata)
        chain_callbacks = [therix_trace_handler] if therix_trace_handler else []
    else:
        therix_trace_handler = config    

    pydantic_prompt = summarizer_config['pydantic_model']
    summarization_type = summarizer_config['summarization_type']
    llm = get_inference_model(inference_model_details.name, inference_model_details.config)

    if summarization_type == 'ABSTRACTIVE':
        map_template = PromptTemplate.from_template(
            """Read the following document and provide a detailed summary of the document.\n{pages}"""
        )
        reduce_template = """The following is a set of summaries:\n{pages}\nTake these and distill them into a final, consolidated summary. Verify each and every detail and make sure that the final summary is accurate and complete."""
        
        map_chain = LLMChain(llm=llm, prompt=map_template)
        reduce_chain = LLMChain(llm=llm, prompt=PromptTemplate.from_template(reduce_template))
        
        combine_documents_chain = StuffDocumentsChain(
            llm_chain=reduce_chain, document_variable_name="pages"
        )

        reduce_documents_chain = ReduceDocumentsChain(
            combine_documents_chain=combine_documents_chain,
            collapse_documents_chain=combine_documents_chain,
            token_max=4000,
        )

        map_reduce_chain = MapReduceDocumentsChain(
            llm_chain=map_chain,
            reduce_documents_chain=reduce_documents_chain,
            document_variable_name="pages",
            return_intermediate_steps=False,
        )

        docs = Document(page_content=text, metadata={"source": "local"})
        text_splitter = CharacterTextSplitter.from_tiktoken_encoder(chunk_size=1000, chunk_overlap=0)
        split_docs = text_splitter.split_documents([docs])

        summary = map_reduce_chain.run(split_docs)
    else:
        summary = text

    parser = JsonOutputParser()
    
    prompt_template = system_prompt.get("system_prompt") if system_prompt and "system_prompt" in system_prompt else system_prompt
    prompt = PromptTemplate(
        template=prompt_template or PROMPT_TEMPLATE[summarization_type],
        input_variables=["context"],
        partial_variables={"response_schema_json": pydantic_prompt},
    )

    if summarization_type == SummarizerTypeMaster.EXTRACTIVE.value:
        chain = prompt | llm | parser
        result = chain.invoke({"context": summary}, config={"callbacks": [therix_trace_handler]})
    else:
        chain = prompt | llm
        result = chain.invoke({"context": summary}, config={"callbacks": [therix_trace_handler]})

    return json.dumps(result) if summarization_type == SummarizerTypeMaster.EXTRACTIVE.value else result.content



from therix.core.pipeline_component import PipelineComponent
from therix.entities.models import ConfigType


class Trace(PipelineComponent):
    def __init__(self, config):
        config["host"] = "https://cloud-api.therix.ai"
        
        super().__init__(ConfigType.TRACE_DETAILS, "LANGFUSE", config)



from importlib import metadata
from therix.core.agent import Agent
from therix.core.inference_models import GroqLlama370b
from therix.core.system_prompt_config import SystemPromptConfig
from therix.core.output_parser import OutputParserWrapper


metadata = {
    "name": "Abhishek Dubey",
}

agent = Agent(name="Side Effects Agent")
(
    agent.add(GroqLlama370b(config={"groq_api_key": "key"}))
    .save()
)
print(agent.id)


agent.invoke("Who is sonu nigam?")



from therix.core.inference_models import GroqLlama31405B, GroqLlama3170B, GroqLlama318B
from therix.core.agent import Agent

agent = Agent(name="My New Published Agent")
(
        agent
        # .add(GroqLlama3170B(
        #     config={"groq_api_key":""}
        # ))
        # .add(GroqLlama318B(
        #     config={"groq_api_key":""}
        # ))
        .add(GroqLlama31405B(
            config={"groq_api_key":""}
        ))
        .save()
    )

print(agent.id)
ans = agent.invoke("Who is sonu nigam?")

print(ans)


import pytest
from therix.utils.pii_filter import pii_filter

@pytest.mark.parametrize("text, entities, expected_output", [
    ("John Doe's email is john@example.com", ["PERSON"], "<strong>John Doe's</strong> email is john@example.com"),
    ("Please contact me at john@example.com", ["EMAIL_ADDRESS"], "Please contact me at <strong>john@example.com</strong>"),
    ("Please call 123-456-7890 for assistance", ["PHONE_NUMBER"], "Please call <strong>123-456-7890</strong> for assistance")
])
def test_pii_filter(text, entities, expected_output):
    assert pii_filter(text, entities) == expected_output


from gc import callbacks
from pickle import NONE
import os
from typing import Any
import json, uuid
from operator import itemgetter
from pathlib import Path
from urllib import response
from langchain.docstore.document import Document
from langchain_community.document_loaders import TextLoader, PyPDFLoader, JSONLoader
from langchain_google_genai import ChatGoogleGenerativeAI, GoogleGenerativeAIEmbeddings
from langchain_openai import OpenAIEmbeddings, ChatOpenAI
from langchain_openai import AzureChatOpenAI
from langchain_groq import ChatGroq
from langchain_core.prompts import ChatPromptTemplate
from langchain_community.document_loaders.youtube import YoutubeLoader
from langchain_openai import AzureOpenAIEmbeddings
from therix.core.JSONLoader import JSONLoader
from langchain_community.embeddings.bedrock import BedrockEmbeddings
import boto3
from langchain_community.llms.bedrock import Bedrock
from therix.core.chat_history.base_chat_history import TherixChatMessageHistory
from therix.core.constants import (
    API_Endpoint,
    DataSourceMaster,
    EmbeddingModelMaster,
    InferenceModelMaster,
)
from therix.core.prompt import Prompt
from therix.core.therix_RAG import VectorStore
from therix.db.session import get_sql_alchemy_url
from langchain_core.prompts import PromptTemplate
from langchain_community.vectorstores.pgvector import PGVector
from langchain.prompts import PromptTemplate
from langchain_core.runnables import RunnablePassthrough
from langchain_core.output_parsers import StrOutputParser
from langchain_core.messages import get_buffer_string
from langchain_core.prompts import format_document
from langchain_core.runnables import RunnableParallel
from langchain_text_splitters import  RecursiveCharacterTextSplitter
from operator import itemgetter
from langchain_experimental.data_anonymizer import PresidioReversibleAnonymizer
from langchain_core.messages.chat import ChatMessage
from therix.services.trace_handler import get_trace_handler
from langchain_core.output_parsers import JsonOutputParser
from langchain_anthropic import ChatAnthropic
import asyncio

def sanitize_text(text):
    # Remove null characters (0x00) from the text
    sanitized_text = text.replace("\x00", "")
    return sanitized_text


def get_loader(file_type, file_path, config):
    if file_type == DataSourceMaster.TEXT:
        return TextLoader(file_path)
    elif file_type == DataSourceMaster.PDF:
        return PyPDFLoader(file_path)
    elif file_type == DataSourceMaster.JSON:
        return JSONLoader(file_path)
    elif file_type == DataSourceMaster.YOUTUBE:
        return YoutubeLoader.from_youtube_url(
            file_path,
            add_video_info=True,
            language=[config["language"], "id"],
            translation="en",
        )
        return PyPDFLoader(file_path)
    elif file_type == DataSourceMaster.WEBSITE:
        return JSONLoader(file_path)
    else:
        raise ValueError(f"Unknown data source type: {file_type}")


def get_embedding_model(embedding_model_name, config):
    if (
        embedding_model_name == EmbeddingModelMaster.OPENAI_TEXT_ADA
        or embedding_model_name == EmbeddingModelMaster.OPENAI_TEXT_EMBEDDING_3_SMALL
        or embedding_model_name == EmbeddingModelMaster.OPENAI_TEXT_EMBEDDING_3_LARGE
    ) and ("api_key" in config):
        return OpenAIEmbeddings(
            openai_api_key=config["api_key"], model=embedding_model_name
        )
    elif (
        embedding_model_name == EmbeddingModelMaster.AZURE_TEXT_ADA
        or embedding_model_name == EmbeddingModelMaster.AZURE_TEXT_EMBEDDING_3_LARGE
        or embedding_model_name == EmbeddingModelMaster.AZURE_TEXT_EMBEDDING_3_SMALL
    ) and ("azure_api_key" in config):
        return AzureOpenAIEmbeddings(
            api_key=config["azure_api_key"],
            model=embedding_model_name,
            azure_endpoint=config["azure_endpoint"],
            openai_api_version=config["openai_api_version"],
            azure_deployment=config["azure_deployment"],
        )

    elif (
        embedding_model_name == EmbeddingModelMaster.BEDROCK_TITAN_EMBEDDING
        or embedding_model_name
        == EmbeddingModelMaster.BEDROCK_TITAN_MULTIMODAL_EMBEDDING
    ) and ("bedrock_aws_session_token" in config):
        bedrock_client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=config["bedrock_aws_access_key_id"],
            aws_secret_access_key=config["bedrock_aws_secret_access_key"],
            aws_session_token=config["bedrock_aws_session_token"],
            region_name=config["bedrock_region_name"],
        )

        return BedrockEmbeddings(
            credentials_profile_name="bedrock-admin",
            client=bedrock_client,
            model_id=embedding_model_name,
            region_name=config["bedrock_region_name"],
        )
        
    elif(
        embedding_model_name == EmbeddingModelMaster.GEMINI_EMBEDDING
    ) and ("google_api_key" in config):
        return GoogleGenerativeAIEmbeddings(
            model=EmbeddingModelMaster.GEMINI_EMBEDDING,
            google_api_key=config["google_api_key"],
        )
    else:
        raise ValueError(f"Unknown embedding model: {embedding_model_name}")
    
    

def get_inference_model(inference_model_name, config):
    if (
        inference_model_name == InferenceModelMaster.OPENAI_GPT_3_5_TURBO
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_3_5_TURBO_INSTRUCT
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4_TURBO_PREVIEW
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4_O
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4_O_MINI
    ) and ("api_key" in config):
        return ChatOpenAI(openai_api_key=config["api_key"], model=inference_model_name)
    elif (
        inference_model_name == InferenceModelMaster.OPENAI_GPT_3_5_TURBO
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_3_5_TURBO_INSTRUCT
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4_TURBO_PREVIEW
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4_O
         or inference_model_name == InferenceModelMaster.OPENAI_GPT_4_O_MINI
    ) and ("azure_api_key" in config):
        return AzureChatOpenAI(
            api_key=config["azure_api_key"],
            model=inference_model_name,
            deployment_name=config["azure_deployment"],
            azure_endpoint=config["azure_endpoint"],
            api_version=config["openai_api_version"],
            temperature=config["temperature"],
        )
    elif (
        inference_model_name == InferenceModelMaster.GROQ_LLM_MIXTRAL_8_7_B
        or inference_model_name == InferenceModelMaster.GROQ_LLM_LLAMA3_70B
        or inference_model_name == InferenceModelMaster.GROQ_LLM_GEMMA7B
        or inference_model_name == InferenceModelMaster.GROQ_LLM_LLAMA3_8B
        or inference_model_name == InferenceModelMaster.GROQ_LLM_LLAMA3_1_8B
        or inference_model_name == InferenceModelMaster.GROQ_LLM_LLAMA3_1_70B
        or inference_model_name == InferenceModelMaster.GROQ_LLM_LLAMA3_1_405B
    ) and ("groq_api_key" in config):
         return ChatGroq(groq_api_key=config["groq_api_key"], model=inference_model_name, temperature=config["temperature"])

    elif (
        inference_model_name == InferenceModelMaster.BEDROCK_TEXT_EXPRES_V1
        or inference_model_name == InferenceModelMaster.BEDROCK_TEXT_LITE_G1
        or inference_model_name == InferenceModelMaster.BEDROCK_TEXT_PREMIER_G1
    ) and ("bedrock_aws_session_token" in config):
        bedrock_client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=config["bedrock_aws_access_key_id"],
            aws_secret_access_key=config["bedrock_aws_secret_access_key"],
            aws_session_token=config["bedrock_aws_session_token"],
            region_name=config["bedrock_region_name"]
        )
        return Bedrock(
            credentials_profile_name="bedrock-admin",
            model_id=inference_model_name,
            client=bedrock_client,
            region_name=config["bedrock_region_name"]
        )
    
    elif(
        inference_model_name == InferenceModelMaster.GOOGLE_GEMINI_PRO
        or inference_model_name == InferenceModelMaster.GOOGLE_GEMINI_1_5_PRO
    ) and ("google_api_key" in config):
        return ChatGoogleGenerativeAI(
            model=inference_model_name,
            google_api_key=config["google_api_key"],
            temperature=config["temperature"], 
            top_p=0.85,
            metadata= {'language': 'en'}
        )
    elif (inference_model_name == InferenceModelMaster.ANTHROPIC_CLAUDE_2_1
        or inference_model_name == InferenceModelMaster.ANTHROPIC_CLAUDE_3_SONNET 
        or inference_model_name == InferenceModelMaster.ANTHROPIC_CLAUDE_3_5_SONNET
        or inference_model_name == InferenceModelMaster.ANTHROPIC_CLAUDE_3_OPUS
        or inference_model_name == InferenceModelMaster.ANTHROPIC_CLAUDE_3_HAIKU
        ) and ("anthropic_api_key" in config):
        return ChatAnthropic(
            model=inference_model_name,
            anthropic_api_key=config["anthropic_api_key"],
            temperature=config["temperature"],
            top_p=0.85
        )
    else:
        raise ValueError(f"Unknown inference model: {inference_model_name}")


async def create_embeddings(data_sources, embedding_model, collection_name):
    if(not os.getenv('THERIX_API_KEY')):
        store = get_vectorstore(embedding_model, collection_name)
    else:
        store = get_vectorstore(embedding_model, collection_name)
        
    config_id = []
    for data_source in data_sources:
        config_id.append(uuid.uuid4())
        # data_source = {'name': 'PDF', 'config': {'files': ['path/to/file', 'path/to/file']}}
        for file_path in data_source.config["files"]:
            loader = get_loader(data_source.name, file_path, data_source.config)
            extra_metadata = [
                ("configId", str(config_id)),
            ]
            pages = loader.load_and_split()
            for page_content in pages:
                for key, value in extra_metadata:
                    page_content.metadata[key] = value
            
            # document.text = mask_data(document.text)
            result=RecursiveCharacterTextSplitter(
                 chunk_size=1000,
                 chunk_overlap=200,
                ).split_documents(pages)
            if(os.getenv('THERIX_API_KEY')):
                await store.add_documents(result)
            else:
                store.add_documents(result)
    # config_id is a list, due to possibility of multiple data_source.
    return {"message" : "embedding created" , "config_id" : config_id}


def get_vectorstore(embedding_model, collection_name):
    embeddings = get_embedding_model(embedding_model.name, embedding_model.config)
    if(not os.getenv('THERIX_API_KEY')):
        store = PGVector(
            collection_name=collection_name,
            connection_string=get_sql_alchemy_url(),
            embedding_function=embeddings,
        )
    else:
        store = VectorStore(
            collection_name=collection_name,
            embedding_function=embeddings,    
        )
    return store


async def async_chat(
    question,
    retriever,
    inference_model,
    embed_model,
    session,
    pipeline_id,
    trace_details=None,
    pii_filter_config=None,
    system_prompt=None,
    output_parser=None,
    trace_api=None,
    db_url=None,
    metadata=None
):
    session_id = session.get('session_id')
    therix_trace_handler = get_trace_handler(trace_details, trace_api, system_prompt, pipeline_id, session_id, metadata)

    chain_callbacks = [therix_trace_handler] if therix_trace_handler else []

    history = TherixChatMessageHistory(
        str(session_id),
        str(pipeline_id),
        db_url or get_sql_alchemy_url(),
        table_name="chat_history",
    )
    message_history =await history.get_message_history(str(session_id))
    chat_history = [
        ChatMessage(role=message["message_role"], content=message["message"])
        for message in message_history
    ]

    if system_prompt:
        template = system_prompt.get("system_prompt") or system_prompt
        if "format_instructions" in template and not output_parser:
            template = template.replace("{format_instructions}", "")

        partial_vars = (
            {"format_instructions": output_parser.get_format_instructions()}
            if output_parser
            else {}
        )
    else:
        template = """
        Answer the question based only on the following context:

        {context}

        Answer the question based on the above context: {question}
        The response should be a valid JSON format with the output_parser structure {format_instructions}.
        """
        partial_vars = (
            {"format_instructions": output_parser.get_format_instructions()}
            if output_parser
            else {}
        )

    if partial_vars:
        ANSWER_PROMPT = PromptTemplate(
            template=template,
            input_variables=["context", "question"],
            partial_variables=partial_vars,
        )
    else:
        ANSWER_PROMPT = ChatPromptTemplate.from_template(template)

    CONDENSE_QUESTION_PROMPT = PromptTemplate.from_template(
        """Given the following conversation and a follow up question, rephrase the follow up question to be a standalone question, in its original language.
        Chat History:
        {chat_history}
        Follow Up Input: {question}
        Standalone question:"""
    )
    
    DEFAULT_DOCUMENT_PROMPT = PromptTemplate.from_template(
        template="{page_content} {source} {page}"
    )

    model = get_inference_model(inference_model.name, inference_model.config)

    if pii_filter_config:
        entities = pii_filter_config.config['entities']
        anonymizer = PresidioReversibleAnonymizer(
            analyzed_fields=entities,
            faker_seed=42
        )

        def _combine_documents(docs):
            doc_strings = [
                format_document(doc, DEFAULT_DOCUMENT_PROMPT)
                for doc in docs
            ]
            return anonymizer.anonymize("\n\n".join(doc_strings))

    else:
        def _combine_documents(docs):
            doc_strings = [
                format_document(doc, DEFAULT_DOCUMENT_PROMPT)
                for doc in docs
            ]
            return "\n\n".join(doc_strings)

    _inputs = RunnableParallel(
        standalone_question=RunnablePassthrough.assign(
            chat_history=lambda x: anonymizer.anonymize(get_buffer_string(x["chat_history"]))
            if pii_filter_config else get_buffer_string(x["chat_history"])
        )
        | CONDENSE_QUESTION_PROMPT
        | model
        | StrOutputParser(),
    )

    if session.get('source') == 'USER':
        _context = {
            "context": itemgetter("standalone_question") | retriever | _combine_documents,
            "question": lambda x: x["standalone_question"],
        }
    else:
        _context = {
            "context": itemgetter("standalone_question") | retriever | _combine_documents,
            "question": lambda x: question,
        }

    conversational_qa_chain = _inputs | _context | ANSWER_PROMPT | model

    config = {}
    if trace_details or trace_api:
        if output_parser:
            config = {"callbacks": [therix_trace_handler]}
        else:
            config = {"callbacks": [therix_trace_handler]}
    if output_parser and not (trace_details or trace_api):
        config = {"callbacks": [therix_trace_handler]}

    result = conversational_qa_chain.invoke(
        {"question": question, "chat_history": chat_history},
        config=config,
    )

    if inference_model.name in [InferenceModelMaster.BEDROCK_TEXT_EXPRES_V1, InferenceModelMaster.BEDROCK_TEXT_LITE_G1]:
        response = str(result)
    else:
        response = json.loads(result.json())["content"]

    if pii_filter_config:
        response = anonymizer.deanonymize(response)

    await history.add_message("user", question, pipeline_id, session_id)
    await history.add_message("system", response, pipeline_id, session_id)
    
    return response


def chat(
    question,
    retriever,
    inference_model,
    embed_model,
    session,
    pipeline_id,
    trace_details=None,
    pii_filter_config=None,
    system_prompt=None,
    output_parser=None,
    trace_api=None,
    metadata=None,
    config = None
):
    session_id = session.get('session_id')
    if not config:
        therix_trace_handler = get_trace_handler(trace_details, trace_api, system_prompt, pipeline_id, session_id, metadata)
        chain_callbacks = [therix_trace_handler] if therix_trace_handler else []
    else :
        therix_trace_handler = config
    history = TherixChatMessageHistory(
        str(session_id),
        str(pipeline_id),
        get_sql_alchemy_url(),
        table_name="chat_history",
    )
    message_history = history.get_message_history(str(session_id))
    chat_history = [
        ChatMessage(role=message["message_role"], content=message["message"])
        for message in message_history
    ]
    
    template = None
    partial_vars = {}

    if system_prompt:
        if "system_prompt" in system_prompt:
            template = system_prompt.get("system_prompt")
        else:
            template = system_prompt     
        if "format_instructions" in template and output_parser is None:
            template = template.replace("{format_instructions}", "")

        if output_parser:
            partial_vars = {"format_instructions": output_parser.get_format_instructions()} 
    else:
        if output_parser:
            template = """
            Answer the question based only on the following context and chat history:

            Context :{context}
            Chat History : {chat_history}

            Question: {question}


            The response should be a valid JSON format with the output_parser structure {format_instructions}.
            """
            partial_vars = {"format_instructions": output_parser.get_format_instructions()}
        else:
            template = """
            Answer the question based only on the following context and chat history:

            Context :{context}
            Chat History : {chat_history}
            question: {question}
            
            """

    if partial_vars:
        ANSWER_PROMPT = PromptTemplate(
            template=template,
            input_variables=["context", "question", "chat_history"],
            partial_variables=partial_vars,
        )
    else:
        ANSWER_PROMPT = PromptTemplate(
            template=template,
            input_variables=["context", "question", "chat_history"]
        )

    CONDENSE_QUESTION_PROMPT = PromptTemplate.from_template(
        """Given the following conversation and a follow up question, rephrase the follow up question to be a standalone question, in its original language.
        Chat History:
        {chat_history}
        Follow Up Input: {question} 
        Standalone question:"""
    )
    
    DEFAULT_DOCUMENT_PROMPT = PromptTemplate.from_template(
        template="{page_content} {source} {page}"
    )

    model = get_inference_model(inference_model.name, inference_model.config)

    if pii_filter_config:
        pii_filter_config = pii_filter_config.config
        entities = pii_filter_config['entities']
        anonymizer = PresidioReversibleAnonymizer(
        analyzed_fields=entities,
        faker_seed=42)

        def _combine_documents(
            docs, document_prompt=DEFAULT_DOCUMENT_PROMPT, document_separator="\n\n"
        ):
            doc_strings = [format_document(doc, document_prompt) for doc in docs]
            return anonymizer.anonymize(document_separator.join(doc_strings))
        
        _inputs = RunnableParallel(
        standalone_question=RunnablePassthrough.assign(
            chat_history=lambda x: anonymizer.anonymize(get_buffer_string(x["chat_history"]))
        )
        | CONDENSE_QUESTION_PROMPT
        | model
        | StrOutputParser(),
        )
    else :
        def _combine_documents(
            docs, document_prompt=DEFAULT_DOCUMENT_PROMPT, document_separator="\n\n"
        ):
            doc_strings = [format_document(doc, document_prompt) for doc in docs]
            return document_separator.join(doc_strings)
            
        _inputs = RunnableParallel(
        standalone_question=RunnablePassthrough.assign(
            chat_history=lambda x: get_buffer_string(x["chat_history"])
        )
        | CONDENSE_QUESTION_PROMPT
        | model
        | StrOutputParser(),
        )
    
    if session.get('source') == 'USER':
        _context = {
            "context": itemgetter("standalone_question") | retriever | _combine_documents,
            "question": lambda x: x["standalone_question"],
            "chat_history": lambda x: chat_history
        }
    else:
        _context = {
            "context": itemgetter("standalone_question") | retriever | _combine_documents,
            "question": lambda x: question,
            "chat_history": lambda x:chat_history
        }

    conversational_qa_chain = _inputs | _context | ANSWER_PROMPT | model

    if((trace_details or trace_api) and output_parser):
        result = conversational_qa_chain.invoke({"question": question, "chat_history": chat_history , "format_instructions" : output_parser.get_format_instructions()},config={"callbacks": [therix_trace_handler]})
    elif((trace_details or trace_api) and not output_parser):
        result = conversational_qa_chain.invoke({"question": question, "chat_history": chat_history},config={"callbacks": [therix_trace_handler]})
    elif(output_parser and not (trace_details or trace_api)):
        result = conversational_qa_chain.invoke({"question": question, "chat_history": chat_history , "format_instructions" : output_parser.get_format_instructions()})
    else:   
        result = conversational_qa_chain.invoke(
            {"question": question, "chat_history": chat_history}
        )
   
    if inference_model.name in [InferenceModelMaster.BEDROCK_TEXT_EXPRES_V1, InferenceModelMaster.BEDROCK_TEXT_LITE_G1]:
        response = str(result)
    else:
        response = json.loads(result.json())["content"]

    if pii_filter_config:
        response = anonymizer.deanonymize(response)

    history.add_message("user", question, pipeline_id, session_id)
    history.add_message("system", response, pipeline_id, session_id)
    
    return response


async def cloud_chat(
    question,
    store,
    inference_model,
    embedding_model,
    session,
    pipeline_id,
    trace_api,
    pii_filter_config,
    system_prompt,
    output_parser,
    metadata,
    trace_details=None,
    top_k=4
):
    session_id = session.get('session_id')
    therix_trace_handler = get_trace_handler(trace_details, trace_api, system_prompt, pipeline_id, session_id, metadata)
    
    if system_prompt:
        system_prompt = system_prompt.get('system_prompt')

    chain_callbacks = [therix_trace_handler] if therix_trace_handler else []
    
    chat_history_endpoint = API_Endpoint.CHAT_HISTORY_ENDPOINT
    
    history = TherixChatMessageHistory(
            session_id=str(session_id),
            pipeline_id=str(pipeline_id),
            api_url=chat_history_endpoint
        )
    message_history_response = await history.async_get_message_history(str(session_id))
    if message_history_response and 'data' in message_history_response and message_history_response['data']:
        message_history = message_history_response['data'][0]
    chat_history = [ChatMessage(role=msg["message_role"], content=msg["message"]) for msg in message_history]
    
    
    
    model = get_inference_model(inference_model.name, inference_model.config)
    
    relevant_documents = await store.get_relevant_documents(query=question, topK=top_k, agent_id = pipeline_id)
    
    context = ""
    for doc in relevant_documents["data"]:
        context += doc["document"]+" "+f"{doc['cmetadata']}"
    
    if(chat_history):
        question = f"""Given the following conversation and a follow up question, answer the question.
            Chat History:
            {chat_history}
            Follow Up Input: {question}
            """
    else:
        question = question
    
    if(output_parser and system_prompt):
        format_instructions = output_parser.get_format_instructions()
        if("format_instructions" in system_prompt):
            prompt = system_prompt
    elif(output_parser and not system_prompt):
        format_instructions = output_parser.get_format_instructions()
        prompt = """You are a skilled professional who understands All kinds of documents. 
                You will be provided with Content and keywords.
                You have Analyze given data.
                Answer the question based only on the following context:
                {context}
                
                standalone question:
                {question}
                
                Response should be in valid JSON format.
                {format_instructions}
                """
                
    elif(system_prompt and not output_parser):
        prompt = system_prompt
    else:
        prompt = """You are a skilled professional who understands All kinds of documents. 
                You will be provided with Content and keywords.
                You have Analyze given data.
                Answer the question based only on the following context:
                {context}
                {question}
                """
                
    if(output_parser):
        prompt = PromptTemplate(
            template=prompt,
            input_variables=["context", "question"],
            partial_variables={"format_instructions": format_instructions}
        )
        chain = prompt | model | output_parser
        
        if(trace_details or trace_api):
            result : Any = chain.invoke({"question": question, "context": context}, config={"callbacks": [therix_trace_handler]})
        else:
            result : Any = chain.invoke({"question": question, "context": context})
        response = json.loads(result.json()) 
        await history.async_add_message("user", question, pipeline_id, session_id)
        await history.async_add_message("system", str(response), pipeline_id, session_id)
    
        return response
    else:
        prompt = PromptTemplate(
            template=prompt,
            input_variables=["context", "question"],
        )
        chain = prompt | model

        if(trace_details or trace_api):
            result : Any = chain.invoke({"question": question, "context": context}, config={"callbacks": [therix_trace_handler]})
        else:
            result : Any = chain.invoke({"question": question, "context": context})

    
        response = json.loads(result.json()) 
        await history.async_add_message("user", question, pipeline_id, session_id)
        await history.async_add_message("system", str(response['content']), pipeline_id, session_id)
        
        return response['content']
    
    # if pii_filter_config:
    #     pii_filter_config = pii_filter_config.config
    #     entities = pii_filter_config['entities']
    #     anonymizer = PresidioReversibleAnonymizer(
    #     analyzed_fields=entities,
    #     faker_seed=42)
        
    

    
    


from typing import List
from pydantic import BaseModel, Field
from therix.core.inference_models import GroqMixtral87bInferenceModel
from therix.core.agent import Agent
import sys
from therix.core.output_parser import OutputParserWrapper
from therix.core.summarizer_output_model import SummarizerOutputModel
from therix.core.summarizer_config import SummarizerConfig
from therix.core.trace import Trace
from therix.utils.summarizer import SummarizerTypeMaster
from therix.core.system_prompt_config import SystemPromptConfig
import os
from therix.services.therix_trace_handler import TherixTraceHandler

GROQ_API_KEY=''

class TopicModel(SummarizerOutputModel):
    mainTopic: str
    subTopic1: str
    subTopic2: str

text = """It’s no secret that vegetables — which are loaded with fiber, vitamins, minerals, and antioxidants — are a must-have in a healthy diet.

Although all vegetables are healthy, several stand out for their supply of nutrients and powerful health benefits.

Here are 14 of the most nutrient-dense veggies available.

1. Spinach
This leafy green tops the chart as one of the most nutrient-dense vegetables.

That’s because 1 cup (30 grams (g)) Trusted Sourceof raw spinach provides 16% of the Daily Value (DV) for vitamin A plus 120% of the DV for vitamin K — all for just 7 calories.

Spinach also boasts antioxidants, which may helpTrusted Source reduce your chance of developing diseases such as cancer.

2. Carrots
Carrots are packed with vitamin A, delivering 119% of the DV in just 1 cup (128 g)Trusted Source. It also contains nutrients like vitamin C and potassium.

They also contain beta-carotene, an antioxidant that providesTrusted Source them with a vibrant orange color. Your body converts it into vitamin A.

One studyTrusted Source of more than 57,000 people associated eating at least 2–4 carrots per week with a 17% lower risk of colorectal cancer in the long run.

A review of 18 studiesTrusted Source also found that carrots may also reduce the chance of developing lung cancer.

3. Broccoli
Just 1 cup (91 g)Trusted Source of raw broccoli provides 77% of the DV for vitamin K, 90% of the DV for vitamin C, and a good amount of folate, manganese, and potassium.

Broccoli is rich in a sulfur-containing plant compound called glucosinolate, as well as its byproduct sulforaphane. It mayTrusted Source be able to help protect against cancer, as well as decreaseTrusted Source inflammation linked to chronic conditions like heart disease.

4. Garlic
GarlicTrusted Source is very nutritious while fairly low on calories, and most people usually consume a small amount as an ingredient in cooking. One clove of garlic only has about 4.5 caloriesTrusted Source. It contains nutrients such as selenium, vitamin C, vitamin B6, and fiber,

It has also been usedTrusted Source as a medicinal plant for millennia. Its main active compound is allicin, which has been shown to aid blood sugar and heart health.

Although further research is needed, test-tube and animal studiesTrusted Source also suggest that allicin has powerful cancer-fighting properties.

5. Brussels sprouts
are a great source of fiber, an important nutrient that supportsTrusted Source bowel regularity, heart health, and blood sugar control. Each servingTrusted Source is also packed with folate, magnesium, and potassium, as well as vitamins A, C, and K.

They also contain kaempferol, an antioxidant that may beTrusted Source particularly effective in preventing cell damage.

Kaempferol has also been shownTrusted Source to have anti-inflammatory and cancer-fighting properties, which may protect against disease.

6. Kale
Only 1 cup (21 g)Trusted Source of raw kale is loaded with potassium, calcium, copper, and vitamins A, B, C, and K.

In one small study, eating kale alongside a high carb meal was more effectiveTrusted Source at preventing blood sugar spikes than eating a high carb meal alone.

Consuming Kale as a powder (made from dried leaves) or drinking its juice has been found in various studies to support decreasingTrusted Source blood pressure, cholesterol, and blood sugar levels. That said, more research is needed to confirm these findings regarding kale juice specifically.

7. Green peas
Peas are a starchy vegetable, which means they have more carbs and calories than non-starchy veggies and may affect blood sugar levels when eaten in large amounts.

Nevertheless, just 1 cup (160 g) Trusted Sourcecontains 9 g of fiber, 9 g of protein, and vitamins A, C, and K, as well as riboflavin, thiamine, niacin, and folate.

Because they’re high in fiber, peas support digestive healthTrusted Source by enhancing the beneficial bacteria in your gut.

Moreover, they are rich in saponins, a group of plant compounds that may help reduceTrusted Source tumor growth and cause cancer cell death.

8. Swiss chard
One cup (36 g)Trusted Source of Swiss chard contains just 7 calories but nearly 1 g of fiber, 1 g of protein, and lots of manganese, magnesium, and vitamins A, C, and K.

It’s also loaded with health-promoting antioxidants and plant compounds, including betalains and flavonoids.

Though more studies are needed, researchTrusted Source has found these compounds may be anti-inflammatory and help reduce the chance of various chronic diseases

9. Beets
Beets are a vibrant, versatile root vegetable that packs fiber, folate, and manganese into each serving with very few caloriesTrusted Source.

They’re also rich in nitrates, which your body converts into nitric oxide — a compound that can help dilate blood vessels. This may help reduce blood pressure and lower the chanceTrusted Source of developing heart disease.

What’s more, beets and their juice have been linkedTrusted Source to improved endurance and athletic performance."""



therix_trace_handler = TherixTraceHandler(
    therix_api_key = os.getenv('THERIX_API_KEY'),
    agent_id = "some-agent-id",
    session_id= "some-session-id",
    metadata = {
        "metadata": "some-metadata"
    } 
)


config = therix_trace_handler.get_handler()



if len(sys.argv) > 1:
    agent = Agent.from_id(sys.argv[1])
    question = sys.argv[2]
    ans = agent.invoke(question)
    print(ans)
else:
    agent = Agent(name="Summarizer Agent with Groq")

    (agent
    .add(GroqMixtral87bInferenceModel(config={"groq_api_key": "GROQ_API_KEY"}))
    .add(SummarizerConfig(SummarizerTypeMaster.EXTRACTIVE,TopicModel))
    .save())

    print(agent.id)


ans = agent.invoke(text, config= config)
print(ans)
        # return ans
    


# ASYNCHRONOUS CALL- EXAMPLE
    # async def call_agent(text):
    #     ans = await agent.async_invoke(text)
    #     print(ans)
    #     return ans

    # asyncio.run(call_agent(text))

    # print(ans)



from langfuse.callback import CallbackHandler
from therix.core.version import get_current_version
import json
import os
from therix.core.constants import API_Endpoint



def get_trace_handler(trace_details=None, trace_api=None, system_prompt=None, pipeline_id=None, session_id=None, metadata=None):
    therix_trace_handler = None
    if trace_details:
        therix_trace_handler = CallbackHandler(
            secret_key="some_secret_key",
            public_key=trace_details['public_key'],
            host=trace_details["host"],
            trace_name=str(pipeline_id),
            session_id=str(session_id),
            version=get_current_version(),
            metadata=metadata
        )
    elif trace_api:
        release = json.dumps({
            "prompt_name": system_prompt.get("prompt_name"),
            "prompt_version": system_prompt.get("prompt_version")
        }) if system_prompt and system_prompt.get("prompt_name") and system_prompt.get("prompt_version") else None

        handler_args = {
            "secret_key" : 'some-secret-key' ,
            "public_key" : os.getenv('THERIX_API_KEY'),
            "host": API_Endpoint.TRACE_INGESTION,
            "trace_name": str(pipeline_id),
            "session_id": str(session_id),
            "version": get_current_version(),
            "metadata": metadata
        }

        if release:
            handler_args["release"] = f"""{release}"""

        therix_trace_handler = CallbackHandler(**handler_args)

    return therix_trace_handler



from therix.core.data_sources import PDFDataSource
from therix.core.embedding_models import (
    BedrockTitanEmbedding,

)
from therix.core.inference_models import Anthropic_Claude_Opus, OpenAIGPT4OMiniInferenceModel
from therix.core.agent import Agent
import sys

from therix.core.trace import Trace

agent = Agent(name="My New Published Agent")
(
        agent.add(OpenAIGPT4OMiniInferenceModel(
            config={"api_key":""}
        ))
        .save()
    )

print(agent.id)
ans = agent.invoke("Who is sonu nigam?")

print(ans)


from therix.core.pipeline_component import PipelineComponent
from langchain_core.output_parsers import JsonOutputParser
from therix.entities.models import ConfigType

class SystemPromptConfig(PipelineComponent):
    def __init__(self, config):
        super().__init__(ConfigType.SYSTEM_PROMPT, 'SYSTEM_PROMPT', config=config)
 


from therix.core.data_sources import PDFDataSource
from therix.core.inference_models import BedrockLiteG1, BedrockTextExpressV1, GroqMixtral87bInferenceModel
from therix.core.embedding_models import BedrockTitanEmbedding
from therix.core.pii_filter_config import PIIFilterConfig
from therix.core.agent import Agent
import sys
from therix.core.trace import Trace

agent = Agent(name="PII Filter Agent")
(agent
    .add(PDFDataSource(config={'files': ['../../test-data/Essay-on-Lata-Mangeshkar-final.pdf']}))
    .add(BedrockTitanEmbedding(config={ "bedrock_aws_access_key_id":"",
                                        "bedrock_aws_secret_access_key" : "",
                                        "bedrock_aws_session_token" : "",
                                        "bedrock_region_name" : "us-east-1"
                                            }))
    .add(BedrockLiteG1(config={ "bedrock_aws_access_key_id":"",
                                "bedrock_aws_secret_access_key" : "",
                                "bedrock_aws_session_token" : "",
                                "bedrock_region_name" : "us-east-1"
                                            }))
    .add(PIIFilterConfig(config={
        'entities': ['PERSON','PHONE_NUMBER','EMAIL_ADDRESS']
    }))
        #   .add(
        #         Trace(
        #         config={
        #             "secret_key": "sk-lf-3207d77e-b681-4d1b-b39a-04d7ee42bca6",
        #             "public_key": "pk-lf-874a92be-a68e-47c2-b129-1fd3f25282d7",
        #             "identifier": "Trying rag with SDK",
        #         }
        #     )
        # )
    .save())

agent.preprocess_data()
print(agent.id)
ans = agent.invoke("Whom is the data about? And what are their personal details?")

print(ans)


import requests
from therix.core.constants import API_Endpoint

class ApiService:
    def __init__(self, therix_api_key):
        if not therix_api_key:
            raise ValueError("therix_api_key is required")
        self.therix_api_key = therix_api_key
        self.base_url = API_Endpoint.BASE_URL
        self.headers = {
            "THERIX_API_KEY": self.therix_api_key,
            "Content-Type": "application/json"
        }

    def _request(self, method, endpoint, payload=None, params=None):
        url = f"{self.base_url}{endpoint}"
        response = requests.request(method, url, json=payload, headers=self.headers, params=params)
        return self._handle_response(response)

    def post(self, endpoint, payload):
        return self._request("POST", endpoint, payload)

    def get(self, endpoint, params=None):
        return self._request("GET", endpoint, params=params)
    
    def patch(self, endpoint, payload):
        return self._request("PATCH", endpoint, payload)

    def _handle_response(self, response):
        if response.status_code not in {200, 201}:
            if(response.status_code == 401):
                print(f"Error: {response.status_code}, Invalid API key")
            elif (response.status_code == 404):    
                print(f"Error: {response.status_code},Data Not Found")
            else:   
                print(f"Error: {response.status_code}, Response: {response.text}")
                response.raise_for_status()
        return response.json()


from typing import Any, List
from wsgiref import headers
from langchain_core.documents import Document
import uuid
import aiohttp
import asyncio
import logging
import json
import os

from therix.core.constants import API_Endpoint
class VectorStore:
    def __init__(self, collection_name, embedding_function):
        self.collection = collection_name
        self.embeddings = embedding_function        
        
    async def add_documents(self, documents: List[Document], **kwargs: Any) -> List[str]:
        async with aiohttp.ClientSession() as session:
            tasks = []
            print('Generating Embeddings ...')
            for doc in documents:
                embeddings = self.embeddings.embed_query(doc.page_content)
        
                payload = {
                    "agent_id": self.collection,
                    "document_id": str(uuid.uuid4()),
                    "document": doc.page_content,
                    "embeddings": embeddings,
                    "cmetadata": doc.metadata
                }
                
                api_endpoint = API_Endpoint.EMBEDDING_ENDPOINT
                
                tasks.append(self.post_to_server(session, api_endpoint, payload))

            responses = await asyncio.gather(*tasks)
            return responses


    async def get_relevant_documents(self, query: str , topK: int, agent_id: str):
        try:
            async with aiohttp.ClientSession() as session: 
                embedded_query = self.embeddings.embed_query(query)
                
                api_endpoint = API_Endpoint.EMBEDDING_ENDPOINT
                
                payload = {
                    "embeddings" : embedded_query,
                    "topK" : topK,
                    "agent_id" : agent_id
                }
                headers = {
                "THERIX_API_KEY": os.getenv('THERIX_API_KEY'),
                "Content-Type": "application/json"
            }
                
                async with session.get(api_endpoint, json=payload, headers=headers) as response:
                    response.raise_for_status()
                    return await response.json()
        except aiohttp.ClientError as e:
            logging.error(e)
            return None
    
    async def post_to_server(self, session, url, payload):
        try:
            headers = {
                "THERIX_API_KEY": os.getenv('THERIX_API_KEY'),
                "Content-Type": "application/json"
            }
            async with session.post(url, json=payload, headers=headers) as response:
                response.raise_for_status()
                return await response.json()
        except aiohttp.ClientError as e:
            logging.error(e)
            return None
    



from pydantic import BaseModel, Field
from typing import List
from langchain.output_parsers import PydanticOutputParser


class OutputParserWrapper:
    @staticmethod
    def parse_output(pydantic_object):
        pydantic_parser = PydanticOutputParser(pydantic_object=pydantic_object)
        return pydantic_parser



from therix.core.constants import InferenceModelMaster
from therix.core.pipeline_component import InferenceModel


class OpenAIGPT35TurboInferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.OPENAI_GPT_3_5_TURBO, config=config)


class OpenAIGPT4InferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.OPENAI_GPT_4, config=config)


class OpenAIGPT4OInferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(
            name=InferenceModelMaster.OPENAI_GPT_4_O, config=config
        )

class OpenAIGPT4OMiniInferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(
            name=InferenceModelMaster.OPENAI_GPT_4_O_MINI, config=config
        )
        
        
class OpenAIGPT4TurboPreviewInferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(
            name=InferenceModelMaster.OPENAI_GPT_4_TURBO_PREVIEW, config=config
        )


class AzureOpenAIGPT3TurboPreviewInferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.AZURE_GPT_3_5_TURBO, config=config)


class AzureOpenAIGPT3TurboInstructnferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(
            name=InferenceModelMaster.AZURE_GPT_3_5_TURBO_INSTRUCT, config=config
        )


class AzureOpenAIGPT4InferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.AZURE_GPT_4, config=config)
        
        
class AzureOpenAIGPT4OInferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.AZURE_GPT_4_O, config=config)


class AzureOpenAIGPT4TurboPreviewInferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(
            name=InferenceModelMaster.AZURE_GPT_4_TURBO_PREVIEW, config=config
        )


class GroqMixtral87bInferenceModel(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(
            name=InferenceModelMaster.GROQ_LLM_MIXTRAL_8_7_B, config=config
        )


class GroqLlama38b(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.GROQ_LLM_LLAMA3_8B, config=config)



class GroqGemma7B(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.GROQ_LLM_GEMMA7B, config=config)


class GroqLlama370b(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.GROQ_LLM_LLAMA3_70B, config=config)


class GroqLlama318B(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.GROQ_LLM_LLAMA3_1_8B, config=config)


class GroqLlama3170B(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.GROQ_LLM_LLAMA3_1_70B, config=config)


class GroqLlama31405B(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.GROQ_LLM_LLAMA3_1_405B, config=config)


class BedrockTextExpressV1(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(
            name=InferenceModelMaster.BEDROCK_TEXT_EXPRES_V1, config=config
        )
        
class BedrockTextPremierG1(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(
            name=InferenceModelMaster.BEDROCK_TEXT_PREMIER_G1, config=config
        )


class BedrockLiteG1(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.BEDROCK_TEXT_LITE_G1, config=config)
        
        
class GeminiPro(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.GOOGLE_GEMINI_PRO, config=config)
        
        
class Gemini_1_5_PRO(InferenceModel):
    def __init__(self, config: dict):
        super().__init__(name=InferenceModelMaster.GOOGLE_GEMINI_1_5_PRO, config=config)
        

class Anthropic_Claude_Opus(InferenceModel):
    def __init__(self, config:dict):
        super().__init__(name=InferenceModelMaster.ANTHROPIC_CLAUDE_3_OPUS , config=config)


class Anthropic_Claude_2_1(InferenceModel):
    def __init__(self, config:dict):
        super().__init__(name=InferenceModelMaster.ANTHROPIC_CLAUDE_2_1 , config=config)
        
        
class Anthropic_Claude_3_Haiku(InferenceModel):
    def __init__(self, config:dict):
        super().__init__(name=InferenceModelMaster.ANTHROPIC_CLAUDE_3_HAIKU , config=config)

class Anthropic_Claude_3_Sonnet(InferenceModel):
    def __init__(self, config:dict):
        super().__init__(name=InferenceModelMaster.ANTHROPIC_CLAUDE_3_SONNET , config=config)
        
class Anthropic_Claude_3_5_Sonnet(InferenceModel):
    def __init__(self, config:dict):
        super().__init__(name=InferenceModelMaster.ANTHROPIC_CLAUDE_3_5_SONNET , config=config)


from therix.core.data_sources import PDFDataSource
from therix.core.embedding_models import (
    GeminiAIEmbeddings,
)
from therix.core.inference_models import (
    GeminiPro,
)

from therix.core.agent import Agent
import sys

from therix.core.trace import Trace

if len(sys.argv) > 1:
    agent = Agent.from_id(sys.argv[1])
    question = sys.argv[2]
    session_id = None

    if len(sys.argv) < 4:
        pass
    else:
        session_id = sys.argv[3]

    ans = agent.invoke(question, session_id)
    print(ans)
else:
    agent = Agent(name="My New Published Agent")
    (
        agent.add(PDFDataSource(config={"files": ["../../test-data/doc.pdf"]}))
        .add(GeminiAIEmbeddings(config={"google_api_key": ""}))
        .add(GeminiPro(config={"google_api_key": ""}))
        .add(
            Trace(
                config={
                    "secret_key": "sk-lf-e62aa7ce-c4c8-4c77-ad7d-9d76dfd96db1",
                    "public_key": "pk-lf-282ad728-c1d6-4247-b6cd-8022198591a9",
                    "identifier": "my own agent",
                }
            )
        )
        .save()
    )

    
    agent.preprocess_data()
    print(agent.id)
    ans = agent.invoke("what this document is about?")

    print(ans)



import asyncio
from typing import List
import json
from pydantic import BaseModel, Field
from therix.core.agent import Agent
import sys
from therix.core.output_parser import OutputParserWrapper

sys_prompt = """Answer the question like a pirate based only on the following context and reply with your capabilities if something is out of context.
    Context: 
    {context}
    Question: {question}
    
    Give the output as follows : {format_instructions}
"""
    
variables = {
    "name": "Nilesh Sanap",
    }
    

if len(sys.argv) > 1:
    agent = Agent.from_id(sys.argv[1])
    question = sys.argv[2]
    session_id = None
    if len(sys.argv) >= 4:
        session_id = sys.argv[3]  
    ans = agent.invoke(agent, question, session_id)
    print(ans)
else:
    class TestDetails(BaseModel):
        name: str = Field(description="Name of the Topic")
        description: str = Field(description="Short description of the Topic")
        citations: str = Field(description="add source of every topic, from where it is generated")
        page: str = Field(description="page number of the topic")


    class OutputParserJSON(BaseModel):
        tests: List[TestDetails] = Field(description="Topic")
        
    agent_id = "d166ebce-c92b-4c6d-8d7f-3be73307ee66"
    # session_id = 'bb74e7da-9730-4422-82ff-e0b711416c66'
    agent = Agent.from_id(agent_id)
    ans = agent.invoke(question="What are the experiments performed in the study?")
    print(ans)



"""added_chat_history_fix

Revision ID: 1e7f3dc15e9c
Revises: b8fe483b5a71
Create Date: 2024-05-06 15:24:26.996166

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = '1e7f3dc15e9c'
down_revision: Union[str, None] = 'b8fe483b5a71'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('chat_history', 'user_id')
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('chat_history', sa.Column('user_id', sa.VARCHAR(), autoincrement=False, nullable=False))
    # ### end Alembic commands ###



from hmac import new
from logging import config
import trace
from typing import Any, List
import therix
from therix.core import embedding_models, pipeline
from therix.utils.summarizer import summarizer, async_summarizer
from ..db.session import SessionLocal
from ..entities.models import ConfigType, Pipeline, PipelineConfiguration
import json
import os
import asyncio
import urllib
from therix.services.web_crawling import crawl_website
from therix.utils.pii_filter import pii_filter
from ..db.session import SessionLocal
from therix.utils.rag import (
    chat,
    cloud_chat,
    create_embeddings,
    get_embedding_model,
    get_vectorstore,
    get_loader,
    async_chat,
)
from langchain_openai import AzureOpenAIEmbeddings
from therix.utils.keyword_search import async_keyword_search, cloud_keyword_search, keyword_search, async_keyword_search
from therix.services.api_service import ApiService
from types import SimpleNamespace
from therix.utils.default_chat import async_invoke_default_chat, invoke_default_chat, async_invoke_default_chat, invoke_default_cloud_chat
from langchain.retrievers.merger_retriever import MergerRetriever
import re

class DictToObject:
    def __init__(self, dictionary):
        for key, value in dictionary.items():
            setattr(self, key, value)

class PipelineService:
    def __init__(self, therix_api_key=None, trace_api_key=None):
        self.db_session = SessionLocal()
        self.trace_api_key = trace_api_key
        self.therix_api_key = therix_api_key
        self.api_service = ApiService(therix_api_key) if therix_api_key else None

    def create_pipeline_with_configurations(self, pipeline_data, configurations_data):
        old_configuration_data = configurations_data

        if self.therix_api_key:
            # Adjust configurations before making the API request
            for config_data in configurations_data:
                if config_data["config_type"] == "INFERENCE_MODEL":
                    config = config_data.get("config", {})
                    config.setdefault("temperature", 0.5)
                    config_data["config"] = config

            # Remove 'pipeline_id' before sending the payload
            configurations_data = [{key: value for key, value in config_data.items() if key != "pipeline_id"} for config_data in configurations_data]

            payload = {
                "name": pipeline_data.get("name"),
                "status": pipeline_data.get("status"),
                "type": pipeline_data.get("type"),
                "agent_configurations": configurations_data,
            }
            response_data = self.api_service.post("agent/", payload)
            return DictToObject(response_data['data'])
        else:
            new_pipeline = Pipeline(**pipeline_data)
            self.db_session.add(new_pipeline)
            self.db_session.flush()

            for config_data in configurations_data:
                config_data["pipeline_id"] = str(new_pipeline.id)  # Convert UUID to string here
                new_config = PipelineConfiguration(**config_data)
                if new_config.config_type == "INFERENCE_MODEL":
                    new_config.config.setdefault("temperature", 0.5)
                self.db_session.add(new_config)
            self.db_session.commit()

            if self.trace_api_key:
                self.api_service = ApiService(therix_api_key=self.trace_api_key)
                old_configuration_data = [{key: value for key, value in config_data.items() if key != "pipeline_id"} for config_data in old_configuration_data]
                payload = {
                    "id": str(new_pipeline.id),  
                    "name": pipeline_data.get("name"),
                    "status": pipeline_data.get("status"),
                    "type": pipeline_data.get("type"),
                    "agent_configurations": old_configuration_data,
                }
                response_data = self.api_service.post("agent/", payload)
                return new_pipeline
            else:
                raise Exception("Trace API key not provided")    

            

    def update_pipeline_configuration(self, pipeline_id, component):
        agent_configuration_id = self.get_pipeline_configurations_by_type(
            pipeline_id, component['config_type']
        )
        if agent_configuration_id:
            self.api_service.patch(f"agent-config/{agent_configuration_id[0].id}", component)

    def get_db_url(self):
        database_url = self.api_service.get("agent/db-url")
        return database_url['data']

    def get_trace_creds(self):
        trace_creds = self.api_service.get("trace-keys")
        return trace_creds['data']

    def get_prompt_by_name(self, prompt_name, variables=None):
        response_data = self.api_service.get("prompts/active", params={"prompt_name": prompt_name})
        prompt_template = response_data['data']['prompt']
        name = response_data['data']['name']
        prompt_version = response_data['data']['version']

        def replace_placeholders(template, variables):
            if not variables:
                return template
            def replacer(match):
                key = match.group(1).strip()
                return variables.get(key, match.group(0))
            return re.sub(r'\{([\w\s]+)\}', replacer, template)
        
        formatted_prompt = replace_placeholders(prompt_template, variables)
        return {
            "prompt_name": name,
            "prompt_version": prompt_version,
            "system_prompt": formatted_prompt
        }

    def publish_pipeline(self, pipeline_data):
        pipeline = self.db_session.query(Pipeline).filter_by(id=pipeline_data.get("id")).first()
        if pipeline:
            pipeline.status = "PUBLISHED"
            self.db_session.commit()
        return pipeline

    def get_pipeline(self, pipeline_id):
        if self.therix_api_key:
            response_data = self.api_service.get(f"agent/{pipeline_id}")
            return DictToObject(response_data['data'])
        else:
            return self.db_session.query(Pipeline).filter_by(id=pipeline_id).first()

    def get_pipeline_configuration(self, pipeline_id):
        if self.therix_api_key:
            response_data = self.api_service.get(f"agent/{pipeline_id}")
            return DictToObject(response_data['data'])
        else:
            return self.db_session.query(PipelineConfiguration).filter_by(pipeline_id=pipeline_id).all()

    def get_pipeline_configurations_by_type(self, pipeline_id, config_type):
        if self.therix_api_key:
            params = {"agent_id": pipeline_id, "config_type": config_type if isinstance(config_type, str) else config_type.value}
            response_data = self.api_service.get("agent-config/", params=params)
            data = response_data.get('data', [])
            if not data:
                return None

            first_index_data = data[0]
            if isinstance(first_index_data, list) and first_index_data:
                return [DictToObject(item) for item in first_index_data]
            else:
                return None
        else:
            return self.db_session.query(PipelineConfiguration).filter_by(pipeline_id=pipeline_id, config_type=config_type).all()

    async def preprocess_data(self, pipeline_id):
        data_sources = self.get_pipeline_configurations_by_type(pipeline_id, "INPUT_SOURCE")
        output_file = None
        if data_sources and hasattr(data_sources[0].config, "website"):
            website_url = data_sources[0].config["website"]
            web_content = crawl_website(website_url)
            domain_name = urllib.parse.urlparse(website_url).netloc
            output_file = f"{domain_name}_data.json"
            with open(output_file, "w") as f:
                json.dump(web_content, f, indent=4)
            data_sources[0].config["files"] = [output_file]
            os.remove(output_file)
        embedding_model = self.get_pipeline_configurations_by_type(pipeline_id, "EMBEDDING_MODEL")
        return await create_embeddings(data_sources, embedding_model[0], str(pipeline_id))



    async def async_invoke_pipeline(
        self, 
        pipeline_id, 
        question, 
        session, 
        trace_details=None, 
        system_prompt=None, 
        db_url=None, 
        trace_api=None, 
        pipeline_ids=None, 
        output_parser=None,
        metadata=None,
        top_k=4
    ):
            async def get_store(pipeline_id):
                embedding_model = await self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.EMBEDDING_MODEL)
                if self.therix_api_key:
                    return await get_vectorstore(embedding_model[0], str(pipeline_id), self.get_db_url())
                else:
                    return await get_vectorstore(embedding_model[0], str(pipeline_id))

            if pipeline_ids:
                combined_retrievers = [await get_store(pid).as_retriever(search_kwargs={"k": top_k}) for pid in pipeline_ids]
                retriever = MergerRetriever(retrievers=combined_retrievers)
            else:
                retriever = (await get_store(pipeline_id)).as_retriever(search_kwargs={"k": top_k})

            inference_model = await self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.INFERENCE_MODEL)

            pii_filter_config = await self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.PII_FILTER)
            pii_filter_config = pii_filter_config[0] if pii_filter_config else None

            result = await async_chat(
                question,
                retriever,
                inference_model[0],
                embedding_model,
                session,
                pipeline_id,
                trace_api=trace_api,
                trace_details=trace_details,
                pii_filter_config=pii_filter_config,
                system_prompt=system_prompt,
                db_url=db_url,
                output_parser=output_parser,
                metadata=metadata
            )

            if pii_filter_config:
                entities = pii_filter_config.config['entities']
                return await pii_filter(result, entities)
            else:
                return result


    async def invoke_pipeline(
        self,
        pipeline_id,
        question,
        session,
        trace_details=None,
        system_prompt=None,
        trace_api=None,
        pipeline_ids=None,
        output_parser=None,
        metadata=None,
        config = None,
        top_k = 4
    ):

        embedding_model = None
        if self.therix_api_key:
            def get_store(pipeline_id):
                embedding_model: Any = self.get_pipeline_configurations_by_type(
                    pipeline_id, ConfigType.EMBEDDING_MODEL
                )
                return get_vectorstore(embedding_model[0], str(pipeline_id))

            if(pipeline_ids):
                store = [get_store(pid) for pid in pipeline_ids]
            else:
                store = get_store(pipeline_id)

            inference_model = self.get_pipeline_configurations_by_type(
                pipeline_id, ConfigType.INFERENCE_MODEL
            )

            pii_filter_config = self.get_pipeline_configurations_by_type(
                pipeline_id, ConfigType.PII_FILTER
            )
            pii_filter_config = pii_filter_config[0] if pii_filter_config else None
            result = await cloud_chat(
                    question,
                    store,
                    inference_model[0],
                    embedding_model,
                    session,
                    pipeline_id,
                    trace_api=trace_api,
                    pii_filter_config=pii_filter_config,
                    system_prompt=system_prompt,
                    output_parser=output_parser,
                    metadata=metadata,
                    trace_details=trace_details,
                    top_k=top_k
                )
            
            if pii_filter_config:
                entities = pii_filter_config.config["entities"]
                return pii_filter(result, entities)
            else:
                return result

        else:

            def get_store(pipeline_id):
                embedding_model: Any = self.get_pipeline_configurations_by_type(
                    pipeline_id, ConfigType.EMBEDDING_MODEL
                )
                return get_vectorstore(
                    embedding_model[0], str(pipeline_id)
                )

            if pipeline_ids:
                combined_retrievers = [
                    get_store(pid).as_retriever(search_kwargs={"k": top_k}) for pid in pipeline_ids
                ]
                retriever = MergerRetriever(retrievers=combined_retrievers)
            else:
                retriever = get_store(pipeline_id).as_retriever(search_kwargs={"k": top_k})

            inference_model = self.get_pipeline_configurations_by_type(
                pipeline_id, ConfigType.INFERENCE_MODEL
            )

            pii_filter_config = self.get_pipeline_configurations_by_type(
                pipeline_id, ConfigType.PII_FILTER
            )
            pii_filter_config = pii_filter_config[0] if pii_filter_config else None

            result = chat(
                question,
                retriever,
                inference_model[0],
                embedding_model,
                session,
                pipeline_id,
                trace_api=trace_api,
                trace_details=trace_details,
                pii_filter_config=pii_filter_config,
                system_prompt=system_prompt,
                output_parser=output_parser,
                metadata=metadata,
                config = config
            )

            if pii_filter_config:
                entities = pii_filter_config.config["entities"]
                return pii_filter(result, entities)
            else:
                return result
   

    
    async def async_invoke_summarizer_pipeline(
    self, 
    pipeline_id, 
    text, 
    session, 
    trace_details=None, 
    trace_api=None, 
    system_prompt=None,
    metadata=None
):
        inference_model = (await self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.INFERENCE_MODEL))[0]
        summarizer_config = (await self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.SUMMARIZER))[0].config

        return await async_summarizer(
            summarizer_config,
            inference_model,
            text,
            trace_details,
            pipeline_id,
            system_prompt,
            trace_api=trace_api,
            session_id=session.get('session_id'),
            config = config,
        )

    def invoke_summarizer_pipeline(
    self, 
    pipeline_id, 
    text, 
    session, 
    trace_details=None, 
    trace_api=None, 
    system_prompt=None,
    metadata=None,
    config = None
):
        inference_model = self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.INFERENCE_MODEL)[0]
        summarizer_config = self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.SUMMARIZER)[0].config

        return summarizer(
            summarizer_config,
            inference_model,
            text,
            trace_details,
            pipeline_id,
            system_prompt,
            trace_api=trace_api,
            session_id=session.get('session_id'),
            metadata=metadata,
            config = config
        )

    
    
    async def async_invoke_default_pipeline(
    self, 
    pipeline_id, 
    session, 
    question, 
    trace_details=None, 
    system_prompt=None, 
    trace_api=None, 
    db_url=None,
    metadata=None,
    config = None
):
        inference_model = (await self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.INFERENCE_MODEL))[0]
        db_url = await self.get_db_url() if self.therix_api_key else db_url

        return await async_invoke_default_chat(
            inference_model_details=inference_model,
            session=session,
            pipeline_id=pipeline_id,
            question=question,
            trace_details=trace_details,
            system_prompt=system_prompt,
            trace_api=trace_api,
            db_url=db_url,
            config=config,
            metadata=metadata
        )

    
    
    async def invoke_default_pipeline(
        self, 
        pipeline_id, 
        session, 
        question, 
        trace_details=None, 
        system_prompt=None, 
        trace_api=None, 
        db_url=None,
        metadata=None,
        config = None
    ):
        inference_model = self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.INFERENCE_MODEL)[0]

        if(not self.therix_api_key):
            return invoke_default_chat(
                inference_model_details=inference_model,
                session=session,
                pipeline_id=pipeline_id,
                question=question,
                trace_details=trace_details,
                system_prompt=system_prompt,
                trace_api=trace_api,
                metadata=metadata,
                config = config
            )
        else:
            return await invoke_default_cloud_chat(
            inference_model_details=inference_model,
            session=session,
            pipeline_id=pipeline_id,
            question=question,
            trace_details=trace_details,
            system_prompt=system_prompt,
            trace_api=trace_api,
            metadata=metadata,
            therix_api_key=self.therix_api_key
        )


    async def async_search_keywords(self, keyword_search_params, top_k=4):
        required_params = ["prompt", "keywords", "output_parser"]
        if not all(param in keyword_search_params for param in required_params):
            return "Request is missing required parameters, please provide all the parameters, i.e. pipeline_id, prompt, keywords, output_parser"

        async def get_store(pipeline_id):
            embedding_model = self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.EMBEDDING_MODEL)[0]
            if self.therix_api_key:
                return await get_vectorstore(embedding_model, str(pipeline_id), self.get_db_url())
            return await get_vectorstore(embedding_model, str(pipeline_id))

        if "pipeline_ids" in keyword_search_params:
            combined_retrievers = [await get_store(pipeline_id).as_retriever(search_kwargs={"k": top_k}) for pipeline_id in keyword_search_params["pipeline_ids"]]
            retriever = MergerRetriever(retrievers=combined_retrievers)
        else:
            pipeline_id = keyword_search_params.get("pipeline_id")
            retriever = await get_store(pipeline_id).as_retriever(search_kwargs={"k": top_k})

        active_pipeline_id = keyword_search_params.get("active_pipeline_id") or \
                            keyword_search_params.get("pipeline_ids", [])[-1] or \
                            keyword_search_params.get("pipeline_id")

        inference_model = self.get_pipeline_configurations_by_type(active_pipeline_id, ConfigType.INFERENCE_MODEL)[0]

        keyword_search_dict = {
            "retriever": retriever,
            "pipeline_id": active_pipeline_id,
            "keywords": keyword_search_params.get("keywords"),
            "output_parser": keyword_search_params.get("output_parser"),
            "prompt": keyword_search_params.get("prompt"),
            "trace_details": keyword_search_params.get("trace_details"),
            "trace_api": keyword_search_params.get("trace_api"),
            "inference_model": inference_model,
            "session": keyword_search_params.get("session"),
        }

        return await async_keyword_search(keyword_search_dict)


    async def search_keywords(self, keyword_search_params, top_k=4):
        # required_params = ["prompt", "keywords", "output_parser"]
        # if not all(param in keyword_search_params for param in required_params):
        #     return "Request is missing required parameters, please provide all the parameters, i.e. pipeline_id, prompt, keywords, output_parser"
        if (self.therix_api_key):
            def get_store(pipeline_id):
                embedding_model = self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.EMBEDDING_MODEL)[0]
                return get_vectorstore(embedding_model, str(pipeline_id))    

            if keyword_search_params["pipeline_ids"]:
                store = [get_store(pipeline_id) for pipeline_id in keyword_search_params["pipeline_ids"]]  
            else:
                pipeline_id = keyword_search_params.get("pipeline_id")
                store = get_store(pipeline_id)

            active_pipeline_id = keyword_search_params.get("active_pipeline_id") or \
                                keyword_search_params.get("pipeline_ids", [])[-1] or \
                                keyword_search_params.get("pipeline_id")

            inference_model = self.get_pipeline_configurations_by_type(active_pipeline_id, ConfigType.INFERENCE_MODEL)[0]

            keyword_search_dict = {
                "store": store,
                "pipeline_id": active_pipeline_id,
                "keywords": keyword_search_params.get("keywords"),
                "output_parser": keyword_search_params.get("output_parser"),
                "prompt": keyword_search_params.get("prompt"),
                "trace_details": keyword_search_params.get("trace_details"),
                "trace_api": keyword_search_params.get("trace_api"),
                "inference_model": inference_model,
                "session": keyword_search_params.get("session"),
            }

            return await cloud_keyword_search(keyword_search_dict)
        else:
            def get_store(pipeline_id):
                embedding_model = self.get_pipeline_configurations_by_type(pipeline_id, ConfigType.EMBEDDING_MODEL)[0]
                return get_vectorstore(embedding_model, str(pipeline_id))
        
            if keyword_search_params['pipeline_ids'] is not None:
                combined_retrievers = [get_store(pipeline_id).as_retriever(search_kwargs={"k": top_k}) for pipeline_id in keyword_search_params["pipeline_ids"]]
                retriever = MergerRetriever(retrievers=combined_retrievers)
            else:
                pipeline_id = keyword_search_params.get("pipeline_id")
                retriever = get_store(pipeline_id).as_retriever(search_kwargs={"k": top_k})

            active_pipeline_id = keyword_search_params.get("active_pipeline_id") or \
                                keyword_search_params.get("pipeline_ids", [])[-1] or \
                                keyword_search_params.get("pipeline_id")

            inference_model = self.get_pipeline_configurations_by_type(active_pipeline_id, ConfigType.INFERENCE_MODEL)[0]

            keyword_search_dict = {
                "retriever": retriever,
                "pipeline_id": active_pipeline_id,
                "keywords": keyword_search_params.get("keywords"),
                "output_parser": keyword_search_params.get("output_parser"),
                "prompt": keyword_search_params.get("prompt"),
                "trace_details": keyword_search_params.get("trace_details"),
                "trace_api": keyword_search_params.get("trace_api"),
                "inference_model": inference_model,
                "session": keyword_search_params.get("session"),
            }

            return keyword_search(keyword_search_dict)
                
                



    def __del__(self):
        
        self.db_session.close()



import uuid
from sqlalchemy.ext.declarative import declarative_base
from therix.core.pipeline_component import PipelineComponent
from therix.entities.models import ConfigType
import logging
from sqlalchemy.dialects.postgresql import UUID
from therix.db.db_manager import DatabaseManager
from typing import Type
from sqlalchemy.engine.base import Engine
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import Session
from typing import (
    Any,
    Type
)



logger = logging.getLogger(__file__)

Base = declarative_base()


db_manager = DatabaseManager()


class CacheConfig(PipelineComponent):
    def __init__(self, config):
        super().__init__(ConfigType.CACHE_CONFIG, 'CACHE_CONFIG', config)       


class FulltextLLMCache(Base): 
    """Postgres table for fulltext-indexed LLM Cache"""

    __tablename__ = "llm_cache_fulltext"
    id = Column(Integer, primary_key=True)
    pipeline_id = Column(UUID(as_uuid=True), nullable=False)
    question = Column(String, nullable=False)
    llm = Column(String, nullable=True)
    response = Column(String)


class TherixCache():


    def __init__(self, engine: Engine, cache_schema: Type[FulltextLLMCache] = FulltextLLMCache):
        self.engine = engine
        self.cache_schema = cache_schema
        self.cache_schema.metadata.create_all(self.engine)


    def lookup(self, question: str, llm_string: str, pipeline_id):
        if not isinstance(pipeline_id, uuid.UUID):
            try:
                pipeline_id = uuid.UUID(pipeline_id)
            except ValueError:
                raise ValueError("pipeline_id is not a valid UUID")
        session = db_manager.get_session()
        query = session.query(FulltextLLMCache).filter(
            FulltextLLMCache.question == question,
            FulltextLLMCache.pipeline_id == pipeline_id
        )
        result = query.first()
        session.close()
        return result
        
    
    
    def update(self, prompt: str, llm_string: str, answer, pipeline_id) -> None:

        if not isinstance(pipeline_id, uuid.UUID):
            try:
                pipeline_id = uuid.UUID(pipeline_id)
            except ValueError:
                raise ValueError("pipeline_id is not a valid UUID")
        session = db_manager.get_session()
        cache_entry = FulltextLLMCache(
            question=prompt,
            llm=llm_string,
            response=answer,
            pipeline_id=pipeline_id
        )
        session.add(cache_entry)
        session.commit()
        session.close()

    def clear(self, **kwargs: Any) -> None:
        """Clear cache."""
        with Session(self.engine) as session:
            session.query(self.cache_schema).delete()
            session.commit()
    



from therix.core.pipeline_component import PipelineComponent
from langchain_core.output_parsers import JsonOutputParser
from therix.entities.models import ConfigType

class SummarizerConfig(PipelineComponent):
    def __init__(self, summarization_type, pydantic_model = None):
        parser = JsonOutputParser(pydantic_object=pydantic_model)
        json_prompt = parser.get_format_instructions()
        config = {
        'pydantic_model' : json_prompt,
        'summarization_type' : summarization_type.value
        }
        super().__init__(ConfigType.SUMMARIZER, 'SUMMARIZER', config)
 


import os
from pickle import FALSE
import trace
import uuid
import warnings
from uuid import uuid4
from sqlalchemy import create_engine
from therix.core.pipeline_component import PipelineComponent
from therix.core.response import ModelResponse
from therix.db.session import get_sql_alchemy_url
from .constants import PipelineTypeMaster
from ..services.pipeline_service import PipelineService
from ..entities.models import ConfigType
import asyncio

warnings.simplefilter('always', DeprecationWarning)
engine = create_engine(get_sql_alchemy_url()) 

class Pipeline:
    def __init__(self, name, pipeline_ids=None, status="PUBLISHED"):
        if self.__class__.__name__ != 'Agent':
            warnings.warn(
                "\033[93mThe 'Pipeline' class is deprecated and will be removed in a future release. Please use the 'Agent' class instead.\033[0m",
                DeprecationWarning,
                stacklevel=2
            )
        self.pipeline_data = {"name": name, "status": status}
        self.components = []
        self.type = None
        self.pipeline_ids = pipeline_ids
        self.added_configs = set()
        self.create_trace= os.getenv("CREATE_TRACE")
        self.therix_api_key = os.getenv("THERIX_API_KEY")
        if not self.therix_api_key:
            print(
                "\033[93m🚨THERIX_API_KEY is not set. You are going to use therix-SDK services without authentication.\033[0m"
            )
        self.trace_api_key = os.getenv("TRACE_API_KEY")
        self.pipeline_service = PipelineService(therix_api_key=self.therix_api_key,trace_api_key=self.trace_api_key)

    @classmethod
    def from_id(cls, pipeline_id):
        pipeline = cls.__new__(cls)
        pipeline.__init__(None)
        pipeline.load(pipeline_id)
        return pipeline
    
    @classmethod
    def merge(cls, pipeline_ids):
        if not pipeline_ids:
            raise ValueError("No pipeline IDs provided for merging")

        merged_pipeline = cls.__new__(cls)
        merged_pipeline.__init__(None, pipeline_ids)

        for pid in pipeline_ids:
            pipeline = cls.from_id(pid)
            merged_pipeline.components.extend(pipeline.components)

        return merged_pipeline

    def validate_configuration(self, component):
        if not isinstance(component, PipelineComponent):
            raise ValueError("component must be an instance of PipelineComponent")
        config_key = component.type.value
        if config_key in self.added_configs and config_key != ConfigType.INPUT_SOURCE.value:
            raise ValueError(f"Configuration '{component.type.value}' is added multiple times.")
        self.added_configs.add(config_key)

    def set_primary(self, pipeline_id):
        self.pipeline_data = self.pipeline_service.get_pipeline(pipeline_id)
        self.id = self.pipeline_data.id
        self.name = self.pipeline_data.name
        return self.pipeline_data

    def add(self, component):
        if not isinstance(component, PipelineComponent):
            raise ValueError("component must be an instance of PipelineComponent")

        component_type_to_pipeline_type = {
            ConfigType.EMBEDDING_MODEL.value: PipelineTypeMaster.RAG.value,
            ConfigType.SUMMARIZER.value: PipelineTypeMaster.SUMMARIZER.value,
        }

        if component.type.value in component_type_to_pipeline_type and self.type is None:
            self.type = component_type_to_pipeline_type[component.type.value]
        elif component.type.value not in component_type_to_pipeline_type:
            pass
        else:
            raise Exception(f"Cannot add {component.type.value}.")

        self.validate_configuration(component)
        self.components.append(component)
        return self

    def add_data_source(self, name, config):
        return self.add(PipelineComponent(ConfigType.INPUT_SOURCE, name, config))

    def add_embedding_model(self, name, config):
        return self.add(PipelineComponent(ConfigType.EMBEDDING_MODEL, name, config))

    def add_inference_model(self, name, config):
        return self.add(PipelineComponent(ConfigType.INFERENCE_MODEL, name, config))

    def add_output_source(self, name, config):
        return self.add(PipelineComponent(ConfigType.OUTPUT_SOURCE, name, config))

    def save(self):
        configurations_data = [
            {"config_type": component.type.value, "name": component.name, "config": component.config}
            for component in self.components
        ]

        if self.type is None:
            if any(config["config_type"] == ConfigType.INFERENCE_MODEL.value for config in configurations_data):
                self.type = PipelineTypeMaster.DEFAULT.value

        self.pipeline_data["type"] = self.type
        self.pipeline_data = self.pipeline_service.create_pipeline_with_configurations(self.pipeline_data, configurations_data)

        self.id = self.pipeline_data.id
        self.name = self.pipeline_data.name
        return self.pipeline_data

    def get_db_url(self):
        return self.pipeline_service.get_db_url()

    def get_prompt_by_name(self, prompt_name):
        return self.pipeline_service.get_prompt_by_name(prompt_name)

    def get_trace_api(self):
        return self.pipeline_service.get_trace_creds()

    def update_configuration(self, pipeline_id, component):
        if not isinstance(component, PipelineComponent):
            raise ValueError("component must be an instance of PipelineComponent")

        config_data = {"config_type": component.type.value, "name": component.name, "config": component.config}
        return self.pipeline_service.update_pipeine_configuration(pipeline_id=pipeline_id, component=config_data)

    def publish(self):
        return self.pipeline_service.publish_pipeline(self.pipeline_data)

    def load(self, pipeline_id):
        self.pipeline_data = self.pipeline_service.get_pipeline(pipeline_id)
        self.id = self.pipeline_data.id
        self.name = self.pipeline_data.name
        self.type = self.pipeline_data.type
        return self.pipeline_data

    def preprocess_data(self):
        async def run_preprocess():
            return await self.pipeline_service.preprocess_data(self.pipeline_data.id)

        return self._invoke_method(run_preprocess)
    
    def _invoke_method(self, method, *args, **kwargs):
        if asyncio.iscoroutinefunction(method):
            return asyncio.run(method(*args, **kwargs))
        else:
            return method(*args, **kwargs)

    def get_configurations(self, pipeline_id):
        inference_model = self.pipeline_service.get_pipeline_configurations_by_type(pipeline_id, ConfigType.INFERENCE_MODEL)
        pipeline_trace_config = self.pipeline_service.get_pipeline_configurations_by_type(pipeline_id, ConfigType.TRACE_DETAILS)
        saved_system_prompt = self.pipeline_service.get_pipeline_configurations_by_type(pipeline_id, ConfigType.SYSTEM_PROMPT)
        pipeline_type = self.pipeline_service.get_pipeline(pipeline_id).type
        if not self.therix_api_key:
            pipeline_type = pipeline_type.value
            
        return inference_model, pipeline_trace_config, saved_system_prompt, pipeline_type

    async def async_invoke(
        self,
        question=None,
        session_id=None,
        keyword_search_params=None,
        dynamic_system_prompt=None,
        output_parser=None,
        variables=None,
        metadata= None
    ):
        async def get_session(session_id):
            return {"session_id": str(uuid.uuid4()), "source": "AUTO-GENERATED"} if session_id is None else {"session_id": session_id, "source": "USER"}

        async def get_system_prompt(saved_system_prompt):
            if self.therix_api_key:
                return {"system_prompt": dynamic_system_prompt} if dynamic_system_prompt else (await self.pipeline_service.get_prompt_by_name(saved_system_prompt[0].config.get('system_prompt'), variables) if saved_system_prompt else None)
            return {"system_prompt": dynamic_system_prompt} if dynamic_system_prompt else (saved_system_prompt[0].config if saved_system_prompt else None)

        async def handle_keyword_search(keyword_search_params, session, trace_details):
            keyword_search_params['active_pipeline_id'] = str(self.pipeline_data.id)
            keyword_search_params["pipeline_ids"] = self.pipeline_ids
            keyword_search_params["trace_details"] = trace_details
            keyword_search_params['session'] = session
            keyword_search_params["trace_api"] = await self.get_trace_api() if self.therix_api_key else None
            return await self.pipeline_service.search_keywords(keyword_search_params)

        async def invoke_pipeline_method(method, *args, **kwargs):
            if pipeline_system_prompt:
                return await method(*args, **kwargs, system_prompt=pipeline_system_prompt, db_url=await self.get_db_url() if self.therix_api_key else None, trace_api=await self.get_trace_api() if self.therix_api_key else None)
            return await method(*args, **kwargs, db_url=await self.get_db_url() if self.therix_api_key else None, trace_api=await self.get_trace_api() if self.therix_api_key and self.create_trace else None)

        if isinstance(self.pipeline_data, dict) and 'id' not in self.pipeline_data:
            if self.pipeline_ids:
                self.set_primary(self.pipeline_ids[-1])

        if question is None and keyword_search_params is None:
            return "Please provide the required parameters to invoke the pipeline"

        session = await get_session(session_id)
        inference_model, pipeline_trace_config, saved_system_prompt, pipeline_type = await self.get_configurations(self.pipeline_data.id)
        pipeline_system_prompt = await get_system_prompt(saved_system_prompt)
        if self.create_trace:
            trace_details = pipeline_trace_config[0].config if pipeline_trace_config else None
        else:
            trace_details = None     

        if pipeline_type == PipelineTypeMaster.RAG.value:
            if keyword_search_params and question is None:
                answer = await handle_keyword_search(keyword_search_params, session, trace_details)
            else:
                output_parser_arg = {"output_parser": output_parser} if output_parser else {}
                pipeline_id = self.pipeline_data.id if isinstance(self.pipeline_data.id, uuid.UUID) else self.pipeline_ids[-1]

                if isinstance(self.pipeline_ids, list) and len(self.pipeline_ids) > 0:
                    self.pipeline_ids.append(str(pipeline_id))
                answer = await invoke_pipeline_method(self.pipeline_service.invoke_pipeline, pipeline_id, question, session, trace_details, pipeline_ids=self.pipeline_ids, **output_parser_arg)

        elif pipeline_type == PipelineTypeMaster.SUMMARIZER.value:
            method = self.pipeline_service.invoke_summarizer_pipeline
            answer = await invoke_pipeline_method(method, self.pipeline_data.id, text=question, trace_details=trace_details, session=session)

        elif pipeline_type == PipelineTypeMaster.DEFAULT.value:
            method = self.pipeline_service.invoke_default_pipeline
            answer = await invoke_pipeline_method(method, self.pipeline_data.id, session, question=question, trace_details=trace_details)

        return ModelResponse(answer, session.get("session_id")).create_response()



    def invoke(
        self,
        question=None,
        session_id=None,
        keyword_search_params=None,
        dynamic_system_prompt=None,
        output_parser=None,
        variables=None,
        metadata=None,
        config = None,
        top_k = None
    ):
        if self.therix_api_key:
            therix_api_key = self.therix_api_key
        else:
            therix_api_key = None

        def get_session(session_id):
            return {"session_id": str(uuid.uuid4()), "source": "AUTO-GENERATED"} if session_id is None else {"session_id": session_id, "source": "USER"}

        def get_system_prompt(saved_system_prompt):
            if self.therix_api_key:
                return {"system_prompt": dynamic_system_prompt} if dynamic_system_prompt else (self.pipeline_service.get_prompt_by_name(saved_system_prompt[0].config.get('system_prompt'), variables) if saved_system_prompt else None)
            return {"system_prompt": dynamic_system_prompt} if dynamic_system_prompt else (saved_system_prompt[0].config if saved_system_prompt else None)

        def handle_keyword_search(keyword_search_params, session, trace_details, top_k):
            keyword_search_params['active_pipeline_id'] = str(self.pipeline_data.id)
            keyword_search_params["pipeline_id"] = self.pipeline_data.id
            keyword_search_params["pipeline_ids"] = self.pipeline_ids
            keyword_search_params["trace_details"] = trace_details
            keyword_search_params['session'] = session
            keyword_search_params["trace_api"] = self.get_trace_api() if self.therix_api_key and self.create_trace else None
            return self.pipeline_service.search_keywords(keyword_search_params, top_k=top_k)

        def invoke_pipeline_method(method, *args, **kwargs):
            if method in [self.pipeline_service.invoke_pipeline, self.pipeline_service.invoke_default_pipeline]:
                kwargs.update({
                    'trace_api': True if self.therix_api_key and self.create_trace else None,
                })
            elif method == self.pipeline_service.invoke_summarizer_pipeline:
                kwargs.update({
                    'trace_api': True if self.therix_api_key and self.create_trace else None
                })
            if asyncio.iscoroutinefunction(method):
                return asyncio.run(method(*args, **kwargs))
            else:
                return method(*args, **kwargs)

        if isinstance(self.pipeline_data, dict) and 'id' not in self.pipeline_data:
            if self.pipeline_ids:
                self.set_primary(self.pipeline_ids[-1])

        if question is None and keyword_search_params is None:
            return "Please provide the required parameters to invoke the pipeline"

        session = get_session(session_id)
        inference_model, pipeline_trace_config, saved_system_prompt, pipeline_type = self.get_configurations(self.pipeline_data.id)
        pipeline_system_prompt = get_system_prompt(saved_system_prompt)
        
        if self.create_trace:
            trace_details = pipeline_trace_config[0].config if pipeline_trace_config else None
        else:
            trace_details = None   

        if pipeline_type == PipelineTypeMaster.RAG.value:
            if keyword_search_params and question is None:
                answer = asyncio.run(handle_keyword_search(keyword_search_params, session, trace_details, top_k))
            else:
                output_parser_arg = {"output_parser": output_parser} if output_parser else {}
                pipeline_id = self.pipeline_data.id
                if isinstance(self.pipeline_ids, list) and len(self.pipeline_ids) > 0:
                    self.pipeline_ids.append(str(pipeline_id))
                answer = invoke_pipeline_method(self.pipeline_service.invoke_pipeline, pipeline_id, question, session, trace_details, pipeline_ids=self.pipeline_ids, **output_parser_arg, system_prompt=pipeline_system_prompt, metadata=metadata, config=config,top_k=top_k)

        elif pipeline_type == PipelineTypeMaster.SUMMARIZER.value: 
            method = self.pipeline_service.invoke_summarizer_pipeline
            answer = invoke_pipeline_method(method, self.pipeline_data.id, text=question, trace_details=trace_details, session=session, system_prompt=pipeline_system_prompt, metadata=metadata, config=config)

        elif pipeline_type == PipelineTypeMaster.DEFAULT.value:
            method = self.pipeline_service.invoke_default_pipeline
            answer = invoke_pipeline_method(method, self.pipeline_data.id, session, question=question, trace_details=trace_details, system_prompt=pipeline_system_prompt, metadata=metadata, config=config)

        return ModelResponse(answer, session.get("session_id")).create_response()



"""changes-type-of-session-id

Revision ID: 3d84bc32c8ee
Revises: f5c4562bcd86
Create Date: 2024-05-21 17:01:09.577057

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '3d84bc32c8ee'
down_revision: Union[str, None] = 'f5c4562bcd86'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    # Create enum type first
    pipeline_type_enum = sa.Enum('RAG', 'SUMMARIZER', 'DEFAULT' , name='pipelinetypemaster')
    pipeline_type_enum.create(op.get_bind())

    # Add the column using the new enum type
    op.add_column('pipelines', sa.Column('type', pipeline_type_enum, nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    # Drop the column first
    op.drop_column('pipelines', 'type')

    # Drop the enum type
    pipeline_type_enum = sa.Enum('RAG', 'SUMMARIZER', name='pipelinetypemaster')
    pipeline_type_enum.drop(op.get_bind())
    # ### end Alembic commands ###


import requests
import time
from selenium import webdriver
from bs4 import BeautifulSoup
import urllib
from selenium.webdriver.chrome.service import Service


def getdata(url):
    try:
        r = requests.get(url)
        return r.text
    except requests.RequestException as e:
        print(f"Request to {url} failed: {e}")
        return None


def crawl_website(website):
    print("Website Crawling Started:", website)
    web_content = []

    options = webdriver.ChromeOptions()
    options.add_argument("--headless")
    options.add_argument("--disable-gpu")
    options.add_argument("--no-sandbox")
    options.add_argument("--disable-dev-shm-usage")

    driver = webdriver.Chrome(options=options)

    def crawl_page(url):
        if url in visited_urls:
            return
        visited_urls.add(url)

        driver.get(url)
        page_source = driver.page_source
        parsed_html = BeautifulSoup(page_source, "html.parser")

        for link in parsed_html.find_all("a", href=True):
            href = link["href"]
            absolute_url = urllib.parse.urljoin(url, href)
            if absolute_url.startswith(website) and absolute_url not in visited_urls:
                crawl_page(absolute_url)

        web_content.append({"url": url, "html_data": getdata(url)})

    visited_urls = set()
    crawl_page(website)

    driver.quit()
    print("Website Crawling Finished:", website)
    return web_content



from enum import Enum
import json
import therix
import asyncio
from therix.core.chat_history.base_chat_history import TherixChatMessageHistory
from therix.core.constants import API_Endpoint, InferenceModelMaster
from therix.db.session import get_sql_alchemy_url
from therix.utils.rag import get_inference_model
from langchain.prompts.chat import ChatPromptTemplate, MessagesPlaceholder
from langfuse.callback import CallbackHandler
from langchain_core.messages.chat import ChatMessage
from therix.services.trace_handler import get_trace_handler


DEFAULT_SYSTEM_PROMPT = "You are a helpful chatbot, you can use your own brain to provide answer to questions by users. The question is: {question}"


async def async_invoke_default_chat(inference_model_details, session, pipeline_id, question, trace_details=None, system_prompt=None, trace_api=None, db_url=None, metadata=None, config = None):
    
    session_id = session.get('session_id')
    if not config:
        therix_trace_handler = get_trace_handler(trace_details, trace_api, system_prompt, pipeline_id, session_id, metadata)
        chain_callbacks = [therix_trace_handler] if therix_trace_handler else []
    else:    
        therix_trace_handler = config
    
    history = TherixChatMessageHistory(
        str(session_id),
        str(pipeline_id),
        db_url or get_sql_alchemy_url(),
        table_name="chat_history",
    )
    message_history = await history.get_message_history(str(session_id))
    chat_history = [ChatMessage(role=msg["message_role"], content=msg["message"]) for msg in message_history]

    model = get_inference_model(inference_model_details.name, inference_model_details.config)
    system_prompt_text = system_prompt.get('system_prompt', DEFAULT_SYSTEM_PROMPT) if system_prompt else DEFAULT_SYSTEM_PROMPT

    ANSWER_PROMPT = ChatPromptTemplate.from_messages([
        ("system", system_prompt_text),
        MessagesPlaceholder("chat_history"),
        ("human", "{question}")
    ])

    chain = ANSWER_PROMPT | model
    chain_input = {"question": question, "chat_history": chat_history}
    config = {"callbacks": [therix_trace_handler]} if therix_trace_handler else {}

    result = await chain.invoke(chain_input, config=config)

    if inference_model_details.name in [InferenceModelMaster.BEDROCK_TEXT_EXPRES_V1, InferenceModelMaster.BEDROCK_TEXT_LITE_G1]:
        response = f'{result}'
    else:
        response = json.loads(result.json())["content"]

    await history.add_message("user", question, pipeline_id, session_id)
    await history.add_message("system", response, pipeline_id, session_id)
    return response


def invoke_default_chat(inference_model_details, session, pipeline_id, question, trace_details=None, system_prompt=None, trace_api=None, metadata=None, config = None):
    session_id = session.get('session_id')

    if not config:
        therix_trace_handler = get_trace_handler(trace_details, trace_api, system_prompt, pipeline_id, session_id, metadata)
        chain_callbacks = [therix_trace_handler] if therix_trace_handler else []
    else:
        therix_trace_handler = config
    history = TherixChatMessageHistory(
            str(session_id),
            str(pipeline_id),
            get_sql_alchemy_url(),
            table_name="chat_history"
    )

    message_history = history.get_message_history(str(session_id))
    chat_history = [ChatMessage(role=msg["message_role"], content=msg["message"]) for msg in message_history]

    model = get_inference_model(inference_model_details.name, inference_model_details.config)
    system_prompt_text = system_prompt.get('system_prompt', DEFAULT_SYSTEM_PROMPT) if system_prompt else DEFAULT_SYSTEM_PROMPT

    ANSWER_PROMPT = ChatPromptTemplate.from_messages([
        ("system", system_prompt_text),
        MessagesPlaceholder("chat_history"),
        ("human", "{question}")
    ])

    chain = ANSWER_PROMPT | model
    chain_input = {"question": question, "chat_history": chat_history}
    config = {"callbacks": [therix_trace_handler]} if therix_trace_handler else {}

    result = chain.invoke(chain_input, config=config)

    if inference_model_details.name in [InferenceModelMaster.BEDROCK_TEXT_EXPRES_V1, InferenceModelMaster.BEDROCK_TEXT_LITE_G1]:
        response = f'{result}'
    else:
        response = json.loads(result.json())["content"]

    history.add_message("user", question, pipeline_id, session_id)
    history.add_message("system", response, pipeline_id, session_id)
    return response


async def invoke_default_cloud_chat(inference_model_details, session, pipeline_id, question, trace_details=None, system_prompt=None, trace_api=None, db_url=None, metadata=None, therix_api_key=None):
    session_id = session.get('session_id')
    therix_trace_handler = get_trace_handler(trace_details, trace_api, system_prompt, pipeline_id, session_id, metadata)

    chain_callbacks = [therix_trace_handler] if therix_trace_handler else []
    
    chat_history_endpoint = API_Endpoint.CHAT_HISTORY_ENDPOINT    
    
    history = TherixChatMessageHistory(
        session_id=str(session_id),
        pipeline_id=str(pipeline_id),
        api_url=chat_history_endpoint
    )

    message_history_response = await history.async_get_message_history(str(session_id))
    if message_history_response and 'data' in message_history_response and message_history_response['data']:
        message_history = message_history_response['data'][0]
    chat_history = [ChatMessage(role=msg["message_role"], content=msg["message"]) for msg in message_history]
    model = get_inference_model(inference_model_details.name, inference_model_details.config)
    system_prompt_text = system_prompt.get('system_prompt', DEFAULT_SYSTEM_PROMPT) if system_prompt else DEFAULT_SYSTEM_PROMPT

    ANSWER_PROMPT = ChatPromptTemplate.from_messages([
        ("system", system_prompt_text),
        MessagesPlaceholder("chat_history"),
        ("human", "{question}")
    ])

    chain = ANSWER_PROMPT | model
    chain_input = {"question": question, "chat_history": chat_history}
    config = {"callbacks": [therix_trace_handler]} if therix_trace_handler else {}

    result = chain.invoke(chain_input, config=config)

    if inference_model_details.name in [InferenceModelMaster.BEDROCK_TEXT_EXPRES_V1, InferenceModelMaster.BEDROCK_TEXT_LITE_G1]:
        response = f'{result}'
    else:
        response = json.loads(result.json())["content"]

    await history.async_add_message("user", question, pipeline_id, session_id)
    await history.async_add_message("system", response, pipeline_id, session_id)
    return response


from therix.core.prompt import Prompt



variables = {
    "history": "User asked about product building",
    "input": "I want to build a new app.",
    "todays_date": "2024-07-10",
    "current_time": "10:00 AM",
    "calendar": "List of booked events",
    "working_hours_start": "9:00 AM",
    "working_hours_end": "6:00 PM"
}

prompt = Prompt.get_prompt("chatbot", variables=variables)
print(prompt)



from presidio_analyzer import AnalyzerEngine
 
def pii_filter(text,entities):
    analyzer = AnalyzerEngine()
    results = analyzer.analyze(text=text, language="en",entities=entities)
    recognized_keywords = []
    keywords = []
    
    def remove_low_scores(result):
        if result.score > 0.2:
            return True
        return False

    recognized_keywords = filter(remove_low_scores, results)
    for item in recognized_keywords:
        if item.score > 0.2:
            start = item.start
            end = item.end
            word = text[start:end]
            keywords.append(word)
            
    keywords = list(set(keywords))
    
    for word in keywords:
        highlighted_word = f"<strong>{word}</strong>"
        text = text.replace(word, highlighted_word)
        
    return text